@@ -2,8 +2,12 @@
 buffer, and return the appropriate TftpPacket object to represent it, via the
 parse() method."""
 
-from TftpShared import *
-from TftpPacketTypes import *
+from __future__ import absolute_import, division, print_function, unicode_literals
+from .TftpShared import *
+from .TftpPacketTypes import *
+import logging
+
+log = logging.getLogger('tftpy.TftpPacketFactory')
 
 class TftpPacketFactory(object):
     """This class generates TftpPacket objects. It is responsible for parsing
@@ -23,9 +27,9 @@ class TftpPacketFactory(object):
         """This method is used to parse an existing datagram into its
         corresponding TftpPacket object. The buffer is the raw bytes off of
         the network."""
-        log.debug("parsing a %d byte packet", len(buffer))
-        (opcode,) = struct.unpack("!H", buffer[:2])
-        log.debug("opcode is %d", opcode)
+        log.debug("parsing a %d byte packet" % len(buffer))
+        (opcode,) = struct.unpack(str("!H"), buffer[:2])
+        log.debug("opcode is %d" % opcode)
         packet = self.__create(opcode)
         packet.buffer = buffer
         return packet.decode()
@@ -33,7 +37,7 @@ class TftpPacketFactory(object):
     def __create(self, opcode):
         """This method returns the appropriate class object corresponding to
         the passed opcode."""
-        tftpassert(self.classes.has_key(opcode),
+        tftpassert(opcode in self.classes,
                    "Unsupported opcode: %d" % opcode)
 
         packet = self.classes[opcode]()
