@@ -1,5 +1,5 @@
 """This module implements all state handling during uploads and downloads, the
-main interface to which being the TftpState base class. 
+main interface to which being the TftpState base class.
 
 The concept is simple. Each context object represents a single upload or
 download, and the state object in the context object represents the current
@@ -8,9 +8,13 @@ the next packet in the transfer, and returns a state object until the transfer
 is complete, at which point it returns None. That is, unless there is a fatal
 error, in which case a TftpException is returned instead."""
 
-from TftpShared import *
-from TftpPacketTypes import *
+from __future__ import absolute_import, division, print_function, unicode_literals
+from .TftpShared import *
+from .TftpPacketTypes import *
 import os
+import logging
+
+log = logging.getLogger('tftpy.TftpStates')
 
 ###############################################################################
 # State classes
@@ -28,7 +32,7 @@ class TftpState(object):
     def handle(self, pkt, raddress, rport):
         """An abstract method for handling a packet. It is expected to return
         a TftpState object, either itself or a new state."""
-        raise NotImplementedError, "Abstract method"
+        raise NotImplementedError("Abstract method")
 
     def handleOACK(self, pkt):
         """This method handles an OACK from the server, syncing any accepted
@@ -42,9 +46,9 @@ class TftpState(object):
                     log.info("    %s = %s" % (key, self.context.options[key]))
             else:
                 log.error("Failed to negotiate options")
-                raise TftpException, "Failed to negotiate options"
+                raise TftpException("Failed to negotiate options")
         else:
-            raise TftpException, "No options found in OACK"
+            raise TftpException("No options found in OACK")
 
     def returnSupportedOptions(self, options):
         """This method takes a requested options list from a client, and
@@ -67,7 +71,7 @@ class TftpState(object):
                     accepted_options[option] = options[option]
             elif option == 'tsize':
                 log.debug("tsize option is set")
-                accepted_options['tsize'] = 1
+                accepted_options['tsize'] = 0
             else:
                 log.info("Dropping unsupported option '%s'" % option)
         log.debug("Returning these accepted options: %s", accepted_options)
@@ -125,9 +129,12 @@ class TftpState(object):
         log.debug("In sendError, being asked to send error %d", errorcode)
         errpkt = TftpPacketERR()
         errpkt.errorcode = errorcode
-        self.context.sock.sendto(errpkt.encode().buffer,
-                                 (self.context.host,
-                                  self.context.tidport))
+        if self.context.tidport == None:
+            log.debug("Error packet received outside session. Discarding")
+        else:
+            self.context.sock.sendto(errpkt.encode().buffer,
+                                     (self.context.host,
+                                      self.context.tidport))
         self.context.last_pkt = errpkt
 
     def sendOACK(self):
@@ -143,7 +150,7 @@ class TftpState(object):
 
     def resendLast(self):
         "Resend the last sent packet due to a timeout."
-        log.warn("Resending packet %s on sessions %s"
+        log.warning("Resending packet %s on sessions %s"
             % (self.context.last_pkt, self))
         self.context.metrics.resent_bytes += len(self.context.last_pkt.buffer)
         self.context.metrics.add_dup(self.context.last_pkt)
@@ -179,10 +186,10 @@ class TftpState(object):
 
         elif pkt.blocknumber < self.context.next_block:
             if pkt.blocknumber == 0:
-                log.warn("There is no block zero!")
+                log.warning("There is no block zero!")
                 self.sendError(TftpErrors.IllegalTftpOp)
-                raise TftpException, "There is no block zero!"
-            log.warn("Dropping duplicate block %d" % pkt.blocknumber)
+                raise TftpException("There is no block zero!")
+            log.warning("Dropping duplicate block %d" % pkt.blocknumber)
             self.context.metrics.add_dup(pkt)
             log.debug("ACKing block %d again, just in case", pkt.blocknumber)
             self.sendACK(pkt.blocknumber)
@@ -192,7 +199,7 @@ class TftpState(object):
             msg = "Whoa! Received future block %d but expected %d" \
                 % (pkt.blocknumber, self.context.next_block)
             log.error(msg)
-            raise TftpException, msg
+            raise TftpException(msg)
 
         # Default is to ack
         return TftpStateExpectDAT(self.context)
@@ -230,9 +237,9 @@ class TftpServerState(TftpState):
 
         # FIXME - only octet mode is supported at this time.
         if pkt.mode != 'octet':
-            self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, \
-                "Only octet transfers are supported at this time."
+            #self.sendError(TftpErrors.IllegalTftpOp)
+            #raise TftpException("Only octet transfers are supported at this time.")
+            log.warning("Received non-octet mode request. I'll reply with binary data.")
 
         # test host/port of client end
         if self.context.host != raddress or self.context.port != rport:
@@ -260,11 +267,10 @@ class TftpServerState(TftpState):
         # begin with a '/' strip it off as otherwise os.path.join will
         # treat it as absolute (regardless of whether it is ntpath or
         # posixpath module
-        if pkt.filename.startswith(self.context.root):
+        if pkt.filename.startswith(self.context.root.encode()):
             full_path = pkt.filename
         else:
-            full_path = os.path.join(
-                self.context.root, pkt.filename.lstrip('/'))
+            full_path = os.path.join(self.context.root, pkt.filename.decode().lstrip('/'))
 
         # Use abspath to eliminate any remaining relative elements
         # (e.g. '..') and ensure that is still within the server's
@@ -274,9 +280,9 @@ class TftpServerState(TftpState):
         if self.full_path.startswith(self.context.root):
             log.info("requested file is in the server root - good")
         else:
-            log.warn("requested file is not within the server root - bad")
+            log.warning("requested file is not within the server root - bad")
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "bad file path"
+            raise TftpException("bad file path")
 
         self.context.file_to_transfer = pkt.filename
 
@@ -299,18 +305,28 @@ class TftpStateServerRecvRRQ(TftpServerState):
         elif self.context.dyn_file_func:
             log.debug("No such file %s but using dyn_file_func", path)
             self.context.fileobj = \
-                self.context.dyn_file_func(self.context.file_to_transfer)
+                self.context.dyn_file_func(self.context.file_to_transfer, raddress=raddress, rport=rport)
 
             if self.context.fileobj is None:
                 log.debug("dyn_file_func returned 'None', treating as "
                           "FileNotFound")
                 self.sendError(TftpErrors.FileNotFound)
-                raise TftpException, "File not found: %s" % path
+                raise TftpException("File not found: %s" % path)
         else:
+            log.warn("File not found: %s", path)
             self.sendError(TftpErrors.FileNotFound)
-            raise TftpException, "File not found: %s" % path
+            raise TftpException("File not found: {}".format(path))
 
         # Options negotiation.
+        if sendoack and self.context.options.has_key('tsize'):
+            # getting the file size for the tsize option. As we handle
+            # file-like objects and not only real files, we use this seeking
+            # method instead of asking the OS
+            self.context.fileobj.seek(0, os.SEEK_END)
+            tsize = str(self.context.fileobj.tell())
+            self.context.fileobj.seek(0, 0)
+            self.context.options['tsize'] = tsize
+
         if sendoack:
             # Note, next_block is 0 here since that's the proper
             # acknowledgement to an OACK.
@@ -348,21 +364,30 @@ class TftpStateServerRecvWRQ(TftpServerState):
                 if os.path.isdir(current):
                     log.debug("%s is already an existing directory", current)
                 else:
-                    os.mkdir(current, 0700)
+                    os.mkdir(current, 0o700)
 
     def handle(self, pkt, raddress, rport):
         "Handle an initial WRQ packet as a server."
         log.debug("In TftpStateServerRecvWRQ.handle")
         sendoack = self.serverInitial(pkt, raddress, rport)
         path = self.full_path
-        log.info("Opening file %s for writing" % path)
-        if os.path.exists(path):
-            # FIXME: correct behavior?
-            log.warn("File %s exists already, overwriting..." % self.context.file_to_transfer)
-        # FIXME: I think we should upload to a temp file and not overwrite the
-        # existing file until the file is successfully uploaded.
-        self.make_subdirs()
-        self.context.fileobj = open(path, "wb")
+        if self.context.upload_open:
+            f = self.context.upload_open(path, self.context)
+            if f is None:
+                self.sendError(TftpErrors.AccessViolation)
+                raise TftpException("Dynamic path %s not permitted" % path)
+            else:
+                self.context.fileobj = f
+        else:
+            log.info("Opening file %s for writing" % path)
+            if os.path.exists(path):
+                # FIXME: correct behavior?
+                log.warning("File %s exists already, overwriting..." % (
+                    self.context.file_to_transfer))
+            # FIXME: I think we should upload to a temp file and not overwrite
+            # the existing file until the file is successfully uploaded.
+            self.make_subdirs()
+            self.context.fileobj = open(path, "wb")
 
         # Options negotiation.
         if sendoack:
@@ -401,8 +426,7 @@ class TftpStateServerStart(TftpState):
                                                                rport)
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, \
-                "Invalid packet to begin up/download: %s" % pkt
+            raise TftpException("Invalid packet to begin up/download: %s" % pkt)
 
 class TftpStateExpectACK(TftpState):
     """This class represents the state of the transfer when a DAT was just
@@ -412,7 +436,7 @@ class TftpStateExpectACK(TftpState):
     def handle(self, pkt, raddress, rport):
         "Handle a packet, hopefully an ACK since we just sent a DAT."
         if isinstance(pkt, TftpPacketACK):
-            log.info("Received ACK for packet %d" % pkt.blocknumber)
+            log.debug("Received ACK for packet %d" % pkt.blocknumber)
             # Is this an ack to the one we just sent?
             if self.context.next_block == pkt.blocknumber:
                 if self.context.pending_complete:
@@ -426,21 +450,20 @@ class TftpStateExpectACK(TftpState):
                     self.context.pending_complete = self.sendDAT()
 
             elif pkt.blocknumber < self.context.next_block:
-                log.warn("Received duplicate ACK for block %d"
+                log.warning("Received duplicate ACK for block %d"
                     % pkt.blocknumber)
                 self.context.metrics.add_dup(pkt)
 
             else:
-                log.warn("Oooh, time warp. Received ACK to packet we "
+                log.warning("Oooh, time warp. Received ACK to packet we "
                          "didn't send yet. Discarding.")
                 self.context.metrics.errors += 1
             return self
         elif isinstance(pkt, TftpPacketERR):
             log.error("Received ERR packet from peer: %s" % str(pkt))
-            raise TftpException, \
-                "Received ERR packet from peer: %s" % str(pkt)
+            raise TftpException("Received ERR packet from peer: %s" % str(pkt))
         else:
-            log.warn("Discarding unsupported packet: %s" % str(pkt))
+            log.warning("Discarding unsupported packet: %s" % str(pkt))
             return self
 
 class TftpStateExpectDAT(TftpState):
@@ -454,19 +477,19 @@ class TftpStateExpectDAT(TftpState):
         elif isinstance(pkt, TftpPacketACK):
             # Umm, we ACK, you don't.
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received ACK from peer when expecting DAT"
+            raise TftpException("Received ACK from peer when expecting DAT")
 
         elif isinstance(pkt, TftpPacketWRQ):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received WRQ from peer when expecting DAT"
+            raise TftpException("Received WRQ from peer when expecting DAT")
 
         elif isinstance(pkt, TftpPacketERR):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received ERR from peer: " + str(pkt)
+            raise TftpException("Received ERR from peer: " + str(pkt))
 
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received unknown packet type from peer: " + str(pkt)
+            raise TftpException("Received unknown packet type from peer: " + str(pkt))
 
 class TftpStateSentWRQ(TftpState):
     """Just sent an WRQ packet for an upload."""
@@ -503,25 +526,25 @@ class TftpStateSentWRQ(TftpState):
                 log.debug("Changing state to TftpStateExpectACK")
                 return TftpStateExpectACK(self.context)
             else:
-                log.warn("Discarding ACK to block %s" % pkt.blocknumber)
+                log.warning("Discarding ACK to block %s" % pkt.blocknumber)
                 log.debug("Still waiting for valid response from server")
                 return self
 
         elif isinstance(pkt, TftpPacketERR):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received ERR from server: " + str(pkt)
+            raise TftpException("Received ERR from server: %s" % pkt)
 
         elif isinstance(pkt, TftpPacketRRQ):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received RRQ from server while in upload"
+            raise TftpException("Received RRQ from server while in upload")
 
         elif isinstance(pkt, TftpPacketDAT):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received DAT from server while in upload"
+            raise TftpException("Received DAT from server while in upload")
 
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received unknown packet type from server: " + str(pkt)
+            raise TftpException("Received unknown packet type from server: %s" % pkt)
 
         # By default, no state change.
         return self
@@ -539,7 +562,7 @@ class TftpStateSentRRQ(TftpState):
             log.info("Received OACK from server")
             try:
                 self.handleOACK(pkt)
-            except TftpException, err:
+            except TftpException as err:
                 log.error("Failed to negotiate options: %s" % str(err))
                 self.sendError(TftpErrors.FailedNegotiation)
                 raise
@@ -564,19 +587,23 @@ class TftpStateSentRRQ(TftpState):
         elif isinstance(pkt, TftpPacketACK):
             # Umm, we ACK, the server doesn't.
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received ACK from server while in download"
+            raise TftpException("Received ACK from server while in download")
 
         elif isinstance(pkt, TftpPacketWRQ):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received WRQ from server while in download"
+            raise TftpException("Received WRQ from server while in download")
 
         elif isinstance(pkt, TftpPacketERR):
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received ERR from server: " + str(pkt)
+            log.debug("Received ERR packet: %s", pkt)
+            if pkt.errorcode == TftpErrors.FileNotFound:
+                raise TftpFileNotFoundError("File not found")
+            else:
+                raise TftpException("Received ERR from server: {}".format(pkt))
 
         else:
             self.sendError(TftpErrors.IllegalTftpOp)
-            raise TftpException, "Received unknown packet type from server: " + str(pkt)
+            raise TftpException("Received unknown packet type from server: %s" % pkt)
 
         # By default, no state change.
         return self
