@@ -8,18 +8,18 @@ As a client of tftpy, this is the only module that you should need to import
 directly. The TftpClient and TftpServer classes can be reached through it.
 """
 
+from __future__ import absolute_import, division, print_function, unicode_literals
 import sys
 
-# Make sure that this is at least Python 2.3
-required_version = (2, 3)
+# Make sure that this is at least Python 2.7
+required_version = (2, 7)
 if sys.version_info < required_version:
-    raise ImportError, "Requires at least Python 2.3"
-
-from tftpy.TftpShared import *
-from tftpy.TftpPacketTypes import *
-from tftpy.TftpPacketFactory import *
-from tftpy.TftpClient import *
-from tftpy.TftpServer import *
-from tftpy.TftpContexts import *
-from tftpy.TftpStates import *
+    raise ImportError("Requires at least Python 2.7")
 
+from .TftpShared import *
+from . import TftpPacketTypes
+from . import TftpPacketFactory
+from .TftpClient import TftpClient
+from .TftpServer import TftpServer
+from . import TftpContexts
+from . import TftpStates
